/** @file
 *
 * @brief Fonctions de gestion du port srie et de la camra
 * @date 2003-2004
 * @author L'quipe Unitec, http://ae.utbm.fr/unitec/
 * @note Code ralis pour la Coupe de France de Robotique 2004
 */

#include "global.h"
#include "x24x_inc.h"

/** Buffer de rception pour la gestion du port srie */
char serial_buffer[64];

/** Position courante dans le buffer de rception du port srie */
int serial_buffer_rx_ptr;

/** Variable permettant de signaler au programme principal la fin de
 * rception du message sur le port srie. Le programme principal la
 * met  0, envoie la commande sur le port srie. La camra va alors
 * envoyer ses informations, qui vont tre lues par le handler
 * d'interruption. Lorsque celui-ci rencontrera ':', c'est  dire la
 * fin de la commande, il mettra cette variable  1 pour signaler au
 * programme principal que la commande est reue
 */
int receptionCommandeTerminee = 0;

/** Comptage des interruptions, pour dbugger */
int nbInterruptionSerie = 0;   

/** Rinitialisation du port srie pour l'envoi d'une nouvelle
 * commande
 *
 */
void serial_init_command(void)
{ 
  int i;

  /* On remet cette valeur a 0 pour pouvoir etre signale de la fin de
     la commande */
  receptionCommandeTerminee = 0;
  
  /* Rinitialisation du buffer de rception */
  serial_buffer_rx_ptr = 0;
  for (i = 0 ; i < 64 ; i++)
  {
  	serial_buffer[i] = 0;
  }
}
                                       
/** Initialisation matrielle du port srie
 *
 */
void serial_port_init(void)
{ 
  int i;
    
  /* Activation de l'horloge pour le port srie */
  SCSR1 = SCSR1 | 0x40;
  
  /* Active les entrees / sorties pour le port srie (0003) */
  MCRA = MCRA | 0x0003;
  
  /* Initialise les indicateurs (reset) */
  SCICTL1 = 0x0;
  
  /* Active les lignes RX et TX (RX Enable + TX Enable + SW Reset) */
  SCICTL1 = 0x23;                                             
  
  /* Format des donnes (1 bit de stop, pas de parite, 8 bits de donnes) */
  SCICCR = 0x7;
  
  /* Configuration de la vitesse (vitesse = 19 200 => BRR = 0x103) */
  SCIHBAUD = 0x0;
  SCILBAUD = 0xBF;
  
  /* Active les interruptions */
  SCICTL2 = 0x2;
  
  SCIPRI = 0x18;        
  
  serial_init_command();
  
  asm("	SETC INTM");
  IFR = 0xFFFF;
  IMR = IMR | 0x1;
  asm("	CLRC INTM");  
}             

/** Envoi d'un caractere sur le port serie 
 *
 * @param c Caractre  envoyer */
 */
static void serial_port_send(unsigned char c)
{                              
  /* Attente du vidage du buffer de transmission */
  while((SCICTL2 & 0x80) == 0);
  
  /* Envoi */
  SCITXBUF = c;
}

/** Reception d'un caractere du port serie 
 *
 * @return Valeur reue par le port srie
 */
static unsigned char serial_port_receive(void)
{              
  /* Attente de la disponibilit d'une donne dans le buffer de rception. 
   * Cette attente n'est plus necessaire, car nous appelons cette fonction 
   * quand l'interruption serie a ete levee, c'est a dire quand une donnee
   * est disponible */
  /* while((SCIRXST & 0x40) == 0); */
  
  /* Rception */
  return (SCIRXBUF & 0xFF);
}

/** Change le mode de la LED de la camra.
 *
 * @param mode Le mode de la LED de la camra (0 = LED teinte, 1 =
 * LED allume, 2 = LED automatique)
 */
void camera_led_change_mode(int mode)
{   
    serial_init_command();    
    mode = mode + '0';
    serial_port_send('L');
    serial_port_send('1');
    serial_port_send(' ');
    serial_port_send(mode);
    serial_port_send('\r');
}

/** Change le "polling mode" de la camra
 *
 * Lorsque le polling mode n'est pas activ la camra envoie 
 * longueur de temps des informations lors d'un suivi de couleur.
 *
 * @param mode Le "polling mode" (0 = Pas de polling, 1 = Du polling)
 */
void camera_pm_mode(int mode)
{   
    serial_init_command();
    mode = mode + '0';
    serial_port_send('P');
    serial_port_send('M');
    serial_port_send(' ');
    serial_port_send('1');
    serial_port_send('\r');
}

/** Rinitialise la camra
* 
*
*/
void camera_reset(void)
{   
    serial_init_command();
    serial_port_send('R');
    serial_port_send('S');
    serial_port_send('\r');
}

/** Dsactive l'auto-gain
 *
 * L'auto-gain permet de corriger la luminosit de l'image en fonction
 * de la lumire. Pour notre cas, l'auto-gain est dsactiv.
 */
void camera_settings(void)
{   
    serial_init_command();
    serial_port_send('C');
    serial_port_send('R');
    serial_port_send(' ');
    serial_port_send('1');
    serial_port_send('9');
    serial_port_send(' ');
    serial_port_send('3');
    serial_port_send('2');
}

/** Envoie une demande de suivi de couleur dans le champ de vision de
 *  la camra
 *
 * @param color La couleur  rechercher. Peut prendre les valeurs
 * RED_COLOR, GREEN_COLOR, BLACK_COLOR.
 */
void camera_track_color(int color)
{
  /* Dfinitions des plages de couleurs correspondantes aux couleurs 
     rechercher */
  char red[] = "TC 149 170 15 17 15 17\r";   /*"TC 110 140 5 25 5 25\r"*/
  char green[] = "TC 70 120 140 190 50 80\r";
  char black[] = "TC 15 40 25 35 8 22\r";
  int i;
  int length;

  /* Rinitialiser le buffer de rception */
  serial_init_command();
  
  /* Si la couleur recherche est le rouge */
  if(color == RED_COLOR)
    {
      length = strlen(red);
      
      for (i = 0; i < length; i++)
	{
	  /* On envoie la commande correspondant  la recherche de la
	     couleur rouge  la camra. La fonction n'envoie qu'un
	     caractre  la fois. On boucle afin d'envoyer toute la
	     commande. */
	  serial_port_send(red[i]);
	}
    }
  /* Si la couleur recherche est le vert */
  else if(color == GREEN_COLOR)
    {
      length = strlen(green);
      
      for (i = 0; i < length; i++)
	{
	  /* On envoie la commande correspondant  la recherche de la
	     couleur verte  la camra. La fonction n'envoie qu'un
	     caractre  la fois. On boucle afin d'envoyer toute la
	     commande. */
	  serial_port_send(green[i]);
	}    
    }
  /* Si la couleur recherche est le noir */
  else if(color == BLACK_COLOR)
    {
      length = strlen(black);
      
      for (i = 0; i < length; i++)
	{
	  /* On envoie la commande correspondant  la recherche de la
	     couleur noire  la camra. La fonction n'envoie qu'un
	     caractre  la fois. On boucle afin d'envoyer toute la
	     commande. */
	  serial_port_send(black[i]);
	}        
    }
}


/** Handler d'interruption pour la rception de donnes sur le port srie.
 *
 */   
void c_int1(void)
{ 
  char val;
  
  asm("	SETC INTM"); 
  IFR = IFR | 0x1;
  
  /* On compte les interruptions (dbuggage) */
  nbInterruptionSerie = nbInterruptionSerie + 1;
  
  /* Rcupration de la valeur sur le port */
  val = serial_port_receive();
  
  /* Si c'est un ':', alors la rception de la commande est termine */
  if(val == ':')
  {
  	receptionCommandeTerminee = 1;
  }
  
  /* Sinon, ajout dans le buffer et incrmentation de la position */
  else
  {  
    serial_buffer[serial_buffer_rx_ptr] = val;
    serial_buffer_rx_ptr = serial_buffer_rx_ptr + 1;
  }
  
  asm("	CLRC INTM");                           
}

/** Vrifie que la commande prcedente s'est bien droule.
 *
 * Cette fonction vrifie que le rsultat de la commande prcdente
 * est correct. Pour cela, on regarde la valeur des 3 premiers
 * octets. Si ils valent ACK, alors la commande s'est bien droule,
 * si ils valent NCK, alors la commande ne s'est pas bien droule.
 * 
 * @return 0 si la commande s'est bien droule, -1 sinon et -2 si le
 * rsultat de la commande est indfini
 */
static int serial_command_successful(void)
{
  if(serial_buffer[0] == 'A' && serial_buffer[1] == 'C' && serial_buffer[2] == 'K')
  {
  	return 0;
  }
  	
  if(serial_buffer[0] == 'N' && serial_buffer[1] == 'C' && serial_buffer[2] == 'K')
  {
  	return -1;
  }
  	
  return -2;
}


/** Attend la fin de la rception du rsultat de la commande.
 *
 * Cette fonction attend que le handler d'interruption de rception du
 * port srie ait reu le rsultat de la commande.
 *
 * @return 0 si la commande s'est bien droule, -1 sinon, -2 si le
 * rsultat est indfini ou -3 si le rsultat n'est pas arriv assez
 * rapidement (timeout).
 */
int serial_command_wait(void)
{        
  int ret;            
  int delay = 7000;
  
  /* On attend que le handler d'interruption nous signale la fin de la
     rception */
  while(receptionCommandeTerminee == 0)
  {
     int i;
     
     /* On attend un peu */
     for (i = 0; i < 100; i++) { i = i + 1 ; i = i - 1; }
     
     /* On dcrmente une variable delay. Ainsi au bout de "delay"
	itrations, on va sortir de la fonction pour grer une sorte
	de timeout afin de ne pas bloquer */
     delay--;
     if(delay == 0)
       {      
	 return -3;
       }
  }
  
  /* Vrification du rsultat de la commande */
  ret = serial_command_successful();
  
  return ret;
}

/** Vrification de la structure d'un paquet de type M
 *
 * Cette fonction vrifie la structure d'un paquet de type M. Les
 * paquets de type M sont ceux renvoys par la camra lors du suivi
 * d'une couleur. Ces paquets contiennent les coordonnes de la zone
 * de la couleur donne, les coordonnes du barycentre ainsi qu'un
 * indice de confiance. Cette fonction va vrifier la structure de
 * manire  ce que la fonction serial_receive_m_command puisse
 * analyser le paquet sans problme.
 *
 * @return 0 si le paquet est bien form, -1 sinon
 */
int serial_check_m_packet()
{  
  int i;
  int spaces = 0;
  
  /* Les octets 0, 1, 2 ont dj t vrifis et valent 'ACK'. L'octet
     3 est un '\r'. On vrifie ici le dbut d'un paquet M  savoir 'M
     ' (M suivi d'un espace) */
  if(serial_buffer[4] != 'M' || serial_buffer[5] != ' ')
  {
    return -1;
  }
  
  /* On vrifie que dans les 64 octets du buffer de rception, on
     trouve la fin du paquet, marque par un '\r' */
  for (i = 4; i < 64 ; i++)
  {
    if(serial_buffer[i] == '\r')
    {
      break;
    }
  }
  
  /* Si on a pas trouv de '\r', on considre le paquet comme
     invalide */
  if(i == 63)
  {
     return -1;
  }
  
  /* On compte le nombre d'espaces du paquet pour vrifier que le bon
     nombre de paramtre a t reu */
  for (i = 4; serial_buffer[i] != '\r' ; i++)
  {
     if(serial_buffer[i] == ' ')
     {
       spaces++;
     }
  }
  
  /* Il faut 8 espaces */
  if(spaces != 8)
  {
     return -1;
  }
  
  return 0;
}

/** Analyse d'un paquet M
 *
 * Cette fonction analyse le contenu d'un paquet de type M. Cette
 * fonction va extraire 3 valeurs intressantes du paquet M : les
 * coordonnes du barycentre de la zone de la couleur donne ainsi que
 * l'indice de confiance.
 *
 * @param x L'adresse  laquelle sera retourne l'abcisse du
 * barycentre
 * @param y L'adresse  laquelle sera retourne l'ordonne du
 * barycentre
 * @param confidence L'adresse  laquelle sera retourne l'indice de
 * confiance
 *
 * @return 0 si l'analyse du paquet a fonctionn, -1 en cas d'erreur
 */
int serial_receive_m_command(int *x, int *y, int *confidence)
{  
   char buf[4];
   int i;
   int j;
   int valx, valy, valconf;
   int ret;
     
   *x = 0;
   *y = 0;
   *confidence = 0;
   
   /* Vrification de la structure du paquet M */
   ret = serial_check_m_packet();
   if(ret < 0)
   {
     return ret;
   }
   
   /* Rinitialisation du petit buffer temporaire */
   for (j = 0; j < 4 ; j++) buf[j] = 0;
   
   /* Rcupration de la coordonne 'x' */
   for (i = 6, j = 0 ; serial_buffer[i] != ' ' ; i++, j++)
   {  
      buf[j] = serial_buffer[i];
   }
   
   /* Passer le ' ' */
   i++;

   /* Conversion en entier */   
   buf[3] = 0;
   valx = atoi(buf);
   *x = valx;
   
   for (j = 0; j < 4 ; j++) buf[j] = 0;
   
   /* Conversion de la coordonne 'y' */   
   for (j = 0 ; serial_buffer[i] != ' ' ; i++, j++)
   {
   	  buf[j] = serial_buffer[i];
   }

   /* Conversion en entier */
   buf[3] = 0;
   valy = atoi(buf);
   *y = valy;
   
   /* Passer le ' ' */
   i++;      
   
   /* Coin gauche x => Inutile */
   for ( ; serial_buffer[i] != ' ' ; i++);
   i++;
   /* Coin gauche y => Inutile */
   for ( ; serial_buffer[i] != ' ' ; i++);
   i++;
   /* Coin droite x => Inutile */
   for ( ; serial_buffer[i] != ' ' ; i++);
   i++;
   /* Coin droite y => Inutile */
   for ( ; serial_buffer[i] != ' ' ; i++);
   i++;   
   /* Nombre de pixels => Inutile */
   for ( ; serial_buffer[i] != ' ' ; i++);
   i++;

   /* Rinitialisation du buffer */
   for (j = 0; j < 4 ; j++) buf[j] = 0;
   
   /* Rcupration de l'indice de confiance */
   for (j = 0 ; ((serial_buffer[i] != '\r') && (serial_buffer[i] != ' ')) ; i++, j++)
   {
      buf[j] = serial_buffer[i];
   }
   
   /* Passer le ' ' */
   i++;

   /* Conversion en entier */   
   buf[3] = 0;
   valconf = atoi(buf);   
   *confidence = valconf;   
   
   return 0;
}
